package FromDualMySQLagent;

#
# Copyright (C) 2010, 2011, 2012 FromDual GmbH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.
# If not, see <http://www.gnu.org/licenses/>.
#

use strict;
use warnings;

use Time::HiRes qw( gettimeofday );

use base 'Exporter';

use constant DBG  => 4;
use constant INFO => 3;
use constant WARN => 2;
use constant ERR  => 1;

our @EXPORT_OK = ('DBG', 'INFO', 'WARN', 'ERR');
our %EXPORT_TAGS = ( stooges => [ 'DBG', 'INFO', 'WARN', 'ERR' ] );

sub sendCachedData
{
  my $pServer = $_[0];
  my $pPort   = $_[1];

  if ( $main::gParameter{'Debug'} >= INFO ) { &log($main::gParameter{'LogFile'}, INFO, '    ' . (caller(0))[3]); }
  if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "Sending cache file to zabbix server."); }

  # File exists and is NOT empty
  if ( -f $main::gParameter{'CacheFile'} ) {

    if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "Cache file found."); }

    my $cmd = "zabbix_sender --zabbix-server $pServer --port $pPort --input-file $main::gParameter{'CacheFile'} --with-timestamps";
    if ( $main::gParameter{'Debug'} == DBG ) {
      $cmd .= ' -vv';
      &log($main::gParameter{'LogFile'}, DBG, "$cmd");
    }

    if ( $main::gParameter{'Debug'} == DBG ) {
      system("$cmd >>$main::gParameter{'LogFile'} 2>&1");
    }
    else {
      system("$cmd >/dev/null 2>&1");
    }
    my $ret = $?;
    my $rc = 1301;
    if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "(ret=$ret / rc=$rc)."); }

    if ( $ret == 0 ) {
      # Do not delete Cache File but just set it back to zero again
      clearCacheFile($main::gParameter{'CacheFile'});
    }
    else {
      &log($main::gParameter{'LogFile'}, ERR, "Load of cache file failed. rc=$rc");
    }
  }
  else {
    if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "No cache file " . $main::gParameter{'CacheFile'} . " found."); }
  }
}

sub clearCacheFile
{
  my $pCacheFile = $_[0];

  if ( $main::gParameter{'Debug'} >= INFO ) { &log($main::gParameter{'LogFile'}, INFO, '      ' . (caller(0))[3]); }

  # Clear Cache File
  my $ret = open(my $fh, '>', $pCacheFile);
  if ( ! defined($ret) ) {
    my $rc = 1311;
    &log($main::gParameter{'LogFile'}, ERR, "  Cannot write to Cache File $pCacheFile: " . $! . "(rc=$rc)");
    return $rc;
  }
  close($fh);
  return 0;
}

sub checkConnectionToZabbixServer
{
  my $pServer = $_[0];
  my $pPort   = $_[1];
  my $pHost   = $_[2];

  if ( $main::gParameter{'Debug'} >= INFO ) { &log($main::gParameter{'LogFile'}, INFO, '      ' . (caller(0))[3]); }

  my $lKey    = 'FromDual.server.check';
  my $lValue  = 1;

  if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "Check connection to zabbix server."); }

  my $cmd = "zabbix_sender --zabbix-server $pServer --port $pPort --host '$pHost' --key $lKey --value '$lValue'";
  if ( $main::gParameter{'Debug'} == DBG ) {
    $cmd .= ' -vv';
    &log($main::gParameter{'LogFile'}, DBG, $cmd);
  }

  if ( $main::gParameter{'Debug'} == DBG ) {
    system("$cmd >>$main::gParameter{'LogFile'} 2>&1");
  }
  else {
    system("$cmd >/dev/null 2>&1");
  }
  my $ret = $?;
  my $rc = 1302;
  if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "(ret=$ret / rc=$rc)."); }
  return $ret;
}

sub readConfigurationFile
{
  my $file = $_[0];

  if ( ! -e "$file" ) {
    print "Configuration file $file does not exist.\n";
    exit(1);
  }

  if ( ! -r "$file" ) {
    print "Cannot read configuration file $file\n";
    exit(2);
  }

  my @config;
  open CONFIG, $file or die $!;
  while ( <CONFIG> ) {
    chomp;                  # remove newlines
    s/^\s*#.*//;            # remove comment
    s/^\s+//;               # remove leading white space
    s/\s+$//;               # remove trailing white space
    next unless length;     # anything left?
    push(@config, $_);
  }
  close(CONFIG);
  return @config;
}

sub getSections
{
  my @config = @_;

  my @sections;
  foreach my $line ( @config ) {

    if ( $line =~ /^\[(.+)\]/ ) {
      push(@sections, $1);
    }
  }

  return @sections;
}

sub readSection
{
  my $config = shift;
  my $section = shift;

  my %section;
  my $in_section = 0;
  foreach my $line ( @$config ) {

    if ( $line =~ /^\[(.+)\]/ ) {
      if ( $1 eq $section ) {
        $in_section = 1;
        next;
      }
      else {
        if ( $in_section == 1 ) {
          last;
        }
        $in_section = 0;
        next;
      }
    }

    if ( $in_section && ( $line =~ /^(\S+)\s*=\s*(.+)\s*$/ ) ) {
      $section{$1} = $2;
    }
  }

  return %section;
}

sub getDefaultParameter
{
  my $config = shift;

  my %hDefaults;

  # Hard coded defaults

  $hDefaults{'ClusterLog'}       = '/var/lib/mysql-cluster/ndb_1_cluster.log';
  $hDefaults{'Debug'}            = WARN;
  $hDefaults{'Disabled'}         = 'false';
  $hDefaults{'LogFile'}          = '/var/log/zabbix/FromDualMySQLagent.log';
  $hDefaults{'Modules'}          = 'mysql myisam process';
  $hDefaults{'MysqlHost'}        = '127.0.0.1';
  $hDefaults{'MysqlPort'}        = '3306';
  $hDefaults{'Password'}         = '';
  $hDefaults{'PidFile'}          = '';
  $hDefaults{'Socket'}           = '/tmp/mysql.sock';
  $hDefaults{'Type'}             = 'mysqld';
  $hDefaults{'Username'}         = 'root';
  $hDefaults{'ZabbixServer'}     = '';
  $hDefaults{'CacheFileBase'}    = '/tmp/FromDualAgentCache';


  $hDefaults{'ZabbixServerPort'} = '10051';

  $hDefaults{'MaaS'}             = 'off';
  $hDefaults{'Hash'}             = '';
  $hDefaults{'Methode'}          = 'http';
  $hDefaults{'Url'}              = 'http://support.fromdual.com/maas/receiver.php';

  # Read default from configuration
  my %default = &readSection(\@$config, 'default');

  foreach my $key ( keys %default ) {

    if ( $default{$key} ne '' ) {
      $hDefaults{$key} = $default{$key};
    }
  }

  return %hDefaults
}

sub getParameter
{
  my $config = shift;
  my $section = shift;

  # The logic is:
  # - Get defaults
  # - Get section parameters
  # - Overwrite defaults by section values
  # - Return defaults as section parameters!

  my %defaults = getDefaultParameter(\@$config);

  if ( $defaults{'Debug'} == DBG ) {
    &log($defaults{'LogFile'}, DBG, "The defaults parameters of this configuration file are:");
    foreach my $p ( keys %defaults ) {

      # This would be possibly a security problem...
      if ( $p eq 'Password' ) {
        &FromDualMySQLagent::log($defaults{'LogFile'}, DBG, "  $p - ********");
      }
      else {
        &FromDualMySQLagent::log($defaults{'LogFile'}, DBG, "  $p - $defaults{$p}");
      }
    }
  }

  my %parameter = readSection(\@$config, $section);

  if ( $defaults{'Debug'} == DBG ) {
    &log($defaults{'LogFile'}, DBG, "The parameters of section $section are:");
    foreach my $p ( keys %parameter ) {

      # This would be possibly a security problem...
      if ( $p eq 'Password' ) {
        &FromDualMySQLagent::log($defaults{'LogFile'}, DBG, "  $p - ********");
      }
      else {
        &FromDualMySQLagent::log($defaults{'LogFile'}, DBG, "  $p - $parameter{$p}");
      }
    }
  }

  foreach my $key ( keys %parameter ) {

    if ( $parameter{$key} ne '' ) {
      $defaults{$key} = $parameter{$key};
    }
  }
  $defaults{'Hostname'} = $section;
  $defaults{'CacheFile'} = $defaults{'CacheFileBase'} . '.' . $section . '.cache';
  return %defaults;
}

sub writeDataToCacheFile
{
  my $values_ref = shift;
  my $list_ref = shift;

  if ( $main::gParameter{'Debug'} >= INFO ) { &log($main::gParameter{'LogFile'}, INFO, '      ' . (caller(0))[3]); }

  # Check if Cache File exists, if not create it
  if ( ! -e $main::gParameter{'CacheFile'} ) {

    my $rc = &FromDualMySQLagent::clearCacheFile($main::gParameter{'CacheFile'});
    if ( $rc != 0 ) {
      my $rc = 1303;
      exit($rc);
    }
  }
  # If ChacheFile exists, check that it is writable
  else {
    if ( ! -w $main::gParameter{'CacheFile'} ) {
      my $rc = 1304;
      &log($main::gParameter{'LogFile'}, ERR, " Cannot write to file $main::gParameter{'CacheFile'} (rc=$rc)!");
      exit($rc);
    }
  }

  my $lDomain = $main::gParameter{'Domain'};
  # some stuff in server module requires no domain
  if ( $lDomain ne '' ) {
    $lDomain .= '.';
  }
  my $lTimestamp = time();

  # Write data to cache file first. We assume this is more efficent than
  # calling zabbix sender zillions of times
  open(DAT, ">>$main::gParameter{'CacheFile'}") || &log($main::gParameter{'LogFile'}, ERR, $! . " Could not open file $main::gCacheFile!");

  foreach my $key ( keys %$values_ref ) {

    # check if key is in ToSend array
    if ( grep $_ eq $key, @$list_ref ) {
      printf(DAT "%s %s %d %s\n", $main::gParameter{'Hostname'}, $lDomain.$key, $lTimestamp, $$values_ref{$key});
    }
  }   # foreach
  close(DAT);
}

sub sendData
{
  my $values_ref = shift;
  my $list_ref = shift;

  my $rc = 0;

  if ( $main::gParameter{'Debug'} >= INFO ) { &log($main::gParameter{'LogFile'}, INFO, '    ' . (caller(0))[3]); }

  &writeDataToCacheFile($values_ref, $list_ref);

  # MPM can be run in 2 modes now: locally or MaaS
  # We run MPM locally
  if ( lc($main::gParameter{'MaaS'}) ne 'on' ) {

    my $rc = &checkConnectionToZabbixServer($main::gParameter{'ZabbixServer'}, $main::gParameter{'ZabbixServerPort'}, $main::gParameter{'Hostname'});

    # Connection to zabbix server seems OK
    if ( $rc == 0 ) {

      if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "Connection to zabbix server seems OK."); }

      # Send cached data
      $rc = sendCachedData($main::gParameter{'ZabbixServer'}, $main::gParameter{'ZabbixServerPort'});
    }
    # No success connecting to zabbix server
    else {
      $rc = 1305;
      &log($main::gParameter{'LogFile'}, WARN, "Connection to zabbix server failed (rc=$rc)!");
    }
  }
  # We run MPM as MaaS
  else {

    if ( lc($main::gParameter{'Methode'}) eq 'http' ) {
        $rc = &uploadData::uploadData();
    }
    else {
      $rc = 1313;
      &log($main::gParameter{'LogFile'}, ERR, "Could NOT upload data to MaaS Server (rc=$rc).");
    }
  }
  return $rc;
}

sub log
{
  my $logfile = shift;
  my $lvl = shift;
  my $str = shift;

  if ( (! defined $logfile) || ($logfile eq '') ) {
    my $rc = 1306;
    print "Logfile is not defined or not specified (rc=$rc).\n";
    exit $rc;
  }

  # Logfile does not exist try to create it
  # This has to be done before the -w check because otherwiese -w would fail the
  # time!
  if ( ! -e $logfile ) {
    open LOG, ">>" . $logfile or die $!;
    close(LOG);
  }

  if ( ! -w $logfile ) {
    my $rc = 1307;
    print "Cannot write to logfile $logfile. Please check permissions (rc=$rc).\n";
    exit $rc;
  }

  my $severity = '';
  if ( $lvl == INFO ) {
    $severity = 'INFO';
  }
  elsif ( $lvl == WARN ) {
    $severity = 'WARN';
  }
  elsif ( $lvl == ERR ) {
    $severity = 'ERR ';
  }
  elsif ( $lvl == DBG ) {
    $severity = 'DBG ';
  }
  else {
    $severity = 'UNKN';
  }

  my $PID = $$;

  open LOG, ">>" . $logfile or die $!;
  my ($seconds, $microseconds) = gettimeofday();
  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($seconds);
  printf(LOG "%5d:%4d-%02d-%02d %02d:%02d:%02d.%03d - %s: %s\n", $PID, $year+1900,$mon+1,$mday,$hour,$min,$sec, $microseconds / 1000, $severity, $str);
  close(LOG);

  # Do a log rotate if necessary if logfile is > 10M
  my $filesize = -s $logfile;
  if ( $filesize > 10*1024*1024 ) {
    my $new = $logfile . '.rotated';
    rename($logfile, $new) or system("mv", $logfile, $new);
  }
}

sub getDatabaseConnection
{
  # Check if perl DBD/mysql is available
  # Should be done on a higher level?
  eval {
    require DBD::mysql;
  };

  my $dbh = undef;
  if ( $@ ) {
    my $rc = 1312;
    &log($main::gParameter{'LogFile'}, WARN, "Warning: No DBD::mysql found. Please install DBD::mysql (rc=$rc).");
    # Should we return an error instead?
    exit($rc);
  }
  else {

    # We do not specify a database, it could be we do not have the
    # privileges for
    $dbh = DBI->connect('DBI:mysql:database=;host=' . $main::gParameter{'MysqlHost'} . ';port=' . $main::gParameter{'MysqlPort'}
                      , $main::gParameter{'Username'}, $main::gParameter{'Password'}
                      , { 'RaiseError' => 0, 'PrintError' => 0 }
                       );

    if ( defined $DBI::err ) {
      &log($main::gParameter{'LogFile'}, ERR, 'DBI connect with database=mysql, host=' . $main::gParameter{'MysqlHost'} . ', port=' . $main::gParameter{'MysqlPort'} . ' and user=' . $main::gParameter{'Username'} . ' failed: ' . $DBI::errstr);
    }
  }
  return $dbh;
}

sub releaseDatabaseConnection()
{
  my $dbh = shift;
  $dbh->disconnect();
}

sub initValues
{
  my $variables_ref = shift;
  my $variables_to_send_ref = shift;

  foreach my $key ( @$variables_to_send_ref ) {
    $$variables_ref{$key} = 0;
  }
}

sub getGlobalVariables
{
  my $dbh = shift;
  my $variables_ref = shift;

  if ( $main::gParameter{'Debug'} >= INFO ) { &FromDualMySQLagent::log($main::gParameter{'LogFile'}, INFO, '    ' . (caller(0))[3]); }

  my $sql = 'SHOW GLOBAL VARIABLES';
  my $sth = $dbh->prepare($sql);
  if ( $sth->execute() ) {
    while ( my $ref = $sth->fetchrow_hashref() ) {
      $$variables_ref{$ref->{'Variable_name'}} = $ref->{'Value'};
    }
    $sth->finish();
  }

  if ( $main::gParameter{'Debug'} == DBG ) {
    &log($main::gParameter{'LogFile'}, DBG, '    Global Variables:');
    foreach my $key ( keys %$variables_ref ) {
      &log($main::gParameter{'LogFile'}, DBG, "      key: $key, value: $$variables_ref{$key}");
    }
  }
}

sub getGlobalStatus
{
  my $dbh = shift;
  my $status_ref = shift;

  if ( $main::gParameter{'Debug'} >= INFO ) { &FromDualMySQLagent::log($main::gParameter{'LogFile'}, INFO, '    ' . (caller(0))[3]); }

  my $sql = 'SHOW /*!50000 GLOBAL */ STATUS';
  my $sth = $dbh->prepare($sql);
  if ( $sth->execute() ) {
    while ( my $ref = $sth->fetchrow_hashref() ) {
      $$status_ref{$ref->{'Variable_name'}} = $ref->{'Value'};
    }
    $sth->finish();
  }

  if ( $main::gParameter{'Debug'} == DBG ) {
    &log($main::gParameter{'LogFile'}, DBG, '    Global Status:');
    foreach my $key ( keys %$status_ref ) {
      &log($main::gParameter{'LogFile'}, DBG, "      key: $key, value: $$status_ref{$key}");
    }
  }
}


sub handleDeadlockMessage
{
  my $lDeadlockMessage = shift;
  my $lTimestamp = time();

  # Cleanup leading and trailing ----
  $lDeadlockMessage =~ s/-+\n//g;

  my $lSendFile = 0;
  # Does an old Deadlock exist?
  my $lDeadlockFile = "/tmp/DeadLockMessage." . $main::gParameter{'Hostname'};
  if ( -f $lDeadlockFile ) {

    # Write temporary Deadlock file
    my $lTmpDeadlockFile = '/tmp/DeadlockMessage.tmp.' . $$;
    `echo '$lDeadlockMessage' > $lTmpDeadlockFile`;
    # Compare with old Deadlock file
    `diff $lTmpDeadlockFile $lDeadlockFile`;
    # Both deadlocks are qual
    if ( $? == 0 ) {
      `rm $lTmpDeadlockFile`;
    }
    # the 2 Deadlocks differ
    else {
      `mv $lTmpDeadlockFile $lDeadlockFile`;
      $lSendFile = 1;
    }

  }
  # no, then write it
  else {
    `echo '$lDeadlockMessage' > $lDeadlockFile`;
    $lSendFile = 1;
  }

  my %hGlobalStatus;

  my @aGlobalStatusToSend = (
    'innodb.Deadlock_new'
  );

  &FromDualMySQLagent::initValues(\%hGlobalStatus, \@aGlobalStatusToSend);

  # Deadlock does not exist yet or is a new deadlock, so send it
  if ( $lSendFile == 1 ) {

    my $rc = &checkConnectionToZabbixServer($main::gParameter{'ZabbixServer'}, $main::gParameter{'ZabbixServerPort'}, $main::gParameter{'Hostname'});

    # Connection to zabbix server seems OK
    if ( $rc == 0 ) {

      if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "Connection to zabbix server seems OK."); }

      my $key = 'innodb.Deadlock_message';

      # This is an exception because with cache file we have not found a way
      # how to send the data with CR/LF
      my $cmd = "zabbix_sender --zabbix-server " . $main::gParameter{'ZabbixServer'} . " --port $main::gParameter{'ZabbixServerPort'} --host '" . $main::gParameter{'Hostname'}. "' --key " . $main::gDomain . $key . " --value '" . $lDeadlockMessage . "'";

      if ( $main::gParameter{'Debug'} == DBG ) {
        $cmd .= ' -vv';
        &log($main::gParameter{'LogFile'}, DBG, $cmd);
      }

      if ( $main::gParameter{'Debug'} == DBG ) {
        system("$cmd >>$main::gParameter{'LogFile'} 2>&1");
      }
      else {
        system("$cmd >/dev/null 2>&1");
      }
      my $ret = $?;
      my $rc = 1308;
      if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "(ret=$ret / rc=$rc)."); }
      # If zabbix_sender fails we should roll back the stuff to get all deadlocks!
    }
    # No success connecting to zabbix server write to Cache File
    else {
      &log($main::gParameter{'LogFile'}, WARN, "Connection to zabbix server failed!");

      # Transform \n into marker 'xxx' because it will cause an error otherwise!
      # We could think about if it is sufficient to send just a deadlock without the timestamp
      # this would cause only the first time a wrong information
      $lDeadlockMessage =~ s/\n/xxx/g;
      array_push(@aGlobalStatusToSend, 'innodb.Deadlock_message');
      $hGlobalStatus{'innodb.Deadlock_message'}  = $lDeadlockMessage;
    }
    $hGlobalStatus{'innodb.Deadlock_new'} = 1;
  }
  # There was NO new Deadlock
  else {
    $hGlobalStatus{'innodb.Deadlock_new'} = 0;
  }
  &writeDataToCacheFile(\%hGlobalStatus, \@aGlobalStatusToSend);
}

sub handleForeignKeyMessage
{
  my $lForeignKeyMessage = shift;
  my $lTimestamp = time();

  # Cleanup leading and trailing ----
  $lForeignKeyMessage =~ s/-+\n//g;

  my $lSendFile = 0;
  # Does an old Foreign Key message exist?
  my $lForeignKeyFile = "/tmp/ForeignKeyMessage." . $main::gParameter{'Hostname'};
  if ( -f $lForeignKeyFile ) {

    # Write temporary Foreign Key file
    my $lTmpForeignKeyFile = '/tmp/ForeignKeyMessage.tmp.' . $$;
    `echo '$lForeignKeyMessage' > $lTmpForeignKeyFile`;
    # Compare with old Foreign Key file
    `diff $lTmpForeignKeyFile $lForeignKeyFile`;
    # Both Foreign Key messages are qual
    if ( $? == 0 ) {
      `rm $lTmpForeignKeyFile`;
    }
    # the 2 Foreign Key messagess differ
    else {
      `mv $lTmpForeignKeyFile $lForeignKeyFile`;
      $lSendFile = 1;
    }

  }
  # no, then write it
  else {
    `echo '$lForeignKeyMessage' > $lForeignKeyFile`;
    $lSendFile = 1;
  }

  my %hGlobalStatus;

  my @aGlobalStatusToSend = (
    'innodb.ForeignKey_new'
  );

  &FromDualMySQLagent::initValues(\%hGlobalStatus, \@aGlobalStatusToSend);

  # Deadlock does not exist yet or is a new deadlock, so send it
  if ( $lSendFile == 1 ) {

    my $rc = &checkConnectionToZabbixServer($main::gParameter{'ZabbixServer'}, $main::gParameter{'ZabbixServerPort'}, $main::gParameter{'Hostname'});

    # Connection to zabbix server seems OK
    if ( $rc == 0 ) {

      if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "Connection to zabbix server seems OK."); }

      my $key = 'innodb.ForeignKey_message';

      # This is an exception because with cache file we have not found a way
      # how to send the data with CR/LF
      my $cmd = "zabbix_sender --zabbix-server " . $main::gParameter{'ZabbixServer'} . " --port $main::gParameter{'ZabbixServerPort'} --host '" . $main::gParameter{'Hostname'}. "' --key " . $main::gDomain . $key . " --value '" . $lForeignKeyMessage . "'";

      if ( $main::gParameter{'Debug'} == DBG ) {
        $cmd .= ' -vv';
        &log($main::gParameter{'LogFile'}, DBG, $cmd);
      }

      if ( $main::gParameter{'Debug'} == DBG ) {
        system("$cmd >>$main::gParameter{'LogFile'} 2>&1");
      }
      else {
        system("$cmd >/dev/null 2>&1");
      }
      my $ret = $?;
      my $rc = 1309;
      if ( $main::gParameter{'Debug'} == DBG ) { &log($main::gParameter{'LogFile'}, DBG, "(ret=$ret / rc=$rc)"); }
      # If zabbix_sender fails we should roll back the stuff to get all foreign key errors!
    }
    # No success connecting to zabbix server write to Cache File
    else {
      &log($main::gParameter{'LogFile'}, WARN, "Connection to zabbix server failed!");

      # Transform \n into marker 'xxx' because it will cause an error otherwise!
      # We could think about if it is sufficient to send just a deadlock without the timestamp
      # this would cause only the first time a wrong information
      $lForeignKeyMessage =~ s/\n/xxx/g;
      array_push(@aGlobalStatusToSend, 'innodb.ForeignKey_message');
      $hGlobalStatus{'innodb.ForeignKey_message'}  = $lForeignKeyMessage;
    }
    $hGlobalStatus{'innodb.ForeignKey_new'} = 1;
  }
  # There was NO new Deadlock
  else {
    $hGlobalStatus{'innodb.ForeignKey_new'} = 0;
  }
  &writeDataToCacheFile(\%hGlobalStatus, \@aGlobalStatusToSend);
}

sub getInnodbStatus
{
  my $dbh = shift;
  my $status_ref = shift;

  if ( $main::gParameter{'Debug'} >= INFO ) { &FromDualMySQLagent::log($main::gParameter{'LogFile'}, INFO, '    ' . (caller(0))[3]); }

  my $sql = 'SHOW /*!50000 ENGINE */ INNODB STATUS';
  my $lInnodbStatus;

  my $sth = $dbh->prepare($sql);
  if ( $sth->execute() ) {
    if ( my $ref = $sth->fetchrow_hashref() ) {
      $lInnodbStatus = $ref->{'Status'};
    }
    $sth->finish();
  }

  # It seems like we do not have access rights!
  # Needs possibly some more investigation here?
  if ( ! defined($lInnodbStatus) || ($lInnodbStatus eq '') ) {
    return 0;
  }

  # Loop over all lines of output
  my $section = '';
  my $lDeadlockMessage   = '';
  my $lForeignKeyMessage = '';
  foreach ( split("\n", $lInnodbStatus) ) {

    # print $_ . "\n";

    if ( $_ =~ /^BACKGROUND THREAD$/ ) {
      $section = 'BACKGROUND THREAD';
      next;
    }
    elsif ( ($_ =~ /^DEAD LOCK ERRORS$/)
         || ($_ =~ /^LATEST DETECTED DEADLOCK$/)
          ) {
      $section = 'DEAD LOCK ERRORS';
      next;
    }
    elsif ( ($_ =~ /^FOREIGN KEY CONSTRAINT ERRORS$/)
         || ($_ =~ /^LATEST FOREIGN KEY ERROR$/)
          ) {
      $section = 'FOREIGN KEY CONSTRAINT ERRORS';
      next;
    }
    elsif ( $_ =~ /^SEMAPHORES$/ ) {
      $section = 'SEMAPHORES';
      next;
    }
    elsif ( $_ =~ /^TRANSACTIONS$/ ) {
      $section = 'TRANSACTIONS';
      next;
    }
    elsif ( $_ =~ /^FILE I\/O$/ ) {
      $section = 'FILE I/O';
      next;
    }
    elsif ( $_ =~ /^INSERT BUFFER AND ADAPTIVE HASH INDEX$/ ) {
      $section = 'INSERT BUFFER AND ADAPTIVE HASH INDEX';
      next;
    }
    elsif ( $_ =~ /^LOG$/ ) {
      $section = 'LOG';
      next;
    }
    elsif ( $_ =~ /^BUFFER POOL AND MEMORY$/ ) {
      $section = 'BUFFER POOL AND MEMORY';
      next;
    }
    elsif ( $_ =~ /^ROW OPERATIONS$/ ) {
      $section = 'ROW OPERATIONS';
      next;
    }

    # --------

    if ( $section eq 'INSERT BUFFER AND ADAPTIVE HASH INDEX' ) {

      # Hash table size 276671, node heap has 810 buffer(s)
      if ( $_ =~ /node heap has (\d+) buffer/ ) {
        $$status_ref{'Innodb_hash_node_size'} = 16384 * $1;
        next;
      }

      # 0.00 hash searches/s, 0.00 non-hash searches/s
      if ( $_ =~ /^(\d+\.\d+)\s+hash searches\/s,\s+(\d+.\d+)\s+non-hash searches\/s/ ) {
        $$status_ref{'Innodb_hash_seaches'} = $1;
        $$status_ref{'Innodb_non_hash_searches'} = $2;
        next;
      }
    }

    if ( $section eq 'SEMAPHORES' ) {

      # Mutex spin waits 27329, rounds 60032, OS waits 929
      if ( $_ =~ /^Mutex spin waits\s+(\d+),\s+rounds\s+(\d+),\s+OS waits\s+(\d+)/ ) {
        $$status_ref{'Innodb_mutex_spin_waits'} = $1;
        $$status_ref{'Innodb_mutex_spin_rounds'} = $2;
        $$status_ref{'Innodb_mutex_os_waits'} = $3;
        next;
      }
    }

    if ( $section eq 'TRANSACTIONS' ) {

      # History list length 177
      if ( $_ =~ /^History list length\s+(\d+)/ ) {
        $$status_ref{'Innodb_trx_history_list_length'} = $1;
        next;
      }
    }

    if ( $section eq 'LOG' ) {

      # Log sequence number 13392740025
      # Log flushed up to   13392740025
      # Pages flushed up to 13392679774
      # Last checkpoint at  13392679671
      if ( $_ =~ /^Log sequence number\s+(\d+)/ ) {
        $$status_ref{'Innodb_log_sequence_number'} = $1;
        next;
      }
      if ( $_ =~ /^Log flushed up to\s+(\d+)/ ) {
        $$status_ref{'Innodb_log_flushed_up_to'} = $1;
        next;
      }
      if ( $_ =~ /^Pages flushed up to\s+(\d+)/ ) {
        $$status_ref{'Innodb_pages_flushed_up_to'} = $1;
        next;
      }
      if ( $_ =~ /^Last checkpoint at\s+(\d+)/ ) {
        $$status_ref{'Innodb_last_checkpoint_at'} = $1;
        next;
      }
    }

    if ( $section eq 'DEAD LOCK ERRORS' ) {
      $lDeadlockMessage .= $_ . "\n";
    }

    if ( $section eq 'FOREIGN KEY CONSTRAINT ERRORS' ) {
      $lForeignKeyMessage .= $_ . "\n";
    }

    if ( $section eq 'ROW OPERATIONS' ) {

      # 0 queries inside InnoDB, 0 queries in queue
      if ( $_ =~ /^(\d+)\s+queries inside InnoDB,\s+(\d+)\s+queries in queue/ ) {
        $$status_ref{'Innodb_queries_inside_innodb'} = $1;
        $$status_ref{'Innodb_queries_in_queue'} = $2;
        next;
      }

      # 1 read views open inside InnoDB
      if ( $_ =~ /^(\d+)\s+read views open inside InnoDB/ ) {
        $$status_ref{'Innodb_open_read_views_inside_innodb'} = $1;
        next;
      }
    }
  }   # loop over InnoDB status

  # Do we have a Deadlock Message?
  if ( $lDeadlockMessage ne '' ) {
    &handleDeadlockMessage($lDeadlockMessage);
  }

  # Do we have a ForeignKey Message?
  if ( $lForeignKeyMessage ne '' ) {
    &handleForeignKeyMessage($lForeignKeyMessage);
  }

  # Get InnoDB Compression Time
  $sql = 'SELECT SUM(compress_time) AS compress_time, SUM(uncompress_time) AS uncompress_time FROM information_schema.INNODB_CMP';

  $sth = $dbh->prepare($sql);
  if ( $sth->execute() ) {
    if ( my $ref = $sth->fetchrow_hashref() ) {
      $$status_ref{'Innodb_compress_time'}   = $ref->{'compress_time'};
      $$status_ref{'Innodb_uncompress_time'} = $ref->{'uncompress_time'};
    }
    $sth->finish();
  }

  # Get InnoDB Row Lock Details and InnoDB Transcation Lock Memory
  $sql = 'SELECT SUM(trx_rows_locked) AS rows_locked, SUM(trx_rows_modified) AS rows_modified, SUM(trx_lock_memory_bytes) AS lock_memory FROM information_schema.INNODB_TRX';

  $sth = $dbh->prepare($sql);
  if ( $sth->execute() ) {
    if ( my $ref = $sth->fetchrow_hashref() ) {
      $$status_ref{'Innodb_rows_locked'}     = defined($ref->{'rows_locked'}) ? $ref->{'rows_locked'} : 0;
      $$status_ref{'Innodb_rows_modified'}   = defined($ref->{'rows_modified'}) ? $ref->{'rows_modified'} : 0;
      $$status_ref{'Innodb_trx_lock_memory'} = defined($ref->{'lock_memory'}) ? $ref->{'lock_memory'} : 0;
    }
    $sth->finish();
  }

  # Get InnoDB Transaction states
  $sql = 'SELECT LOWER(REPLACE(trx_state, " ", "_")) AS state, count(*) AS cnt from information_schema.INNODB_TRX GROUP BY state';

  $sth = $dbh->prepare($sql);
  if ( $sth->execute() ) {
    if ( my $ref = $sth->fetchrow_hashref() ) {
      $$status_ref{'Innodb_trx_' . $ref->{'state'}} = $ref->{'cnt'};
    }
    $sth->finish();
  }
}

sub setAgentLock
{
  my $pLogLevel = $_[0];
  my $pLogFile  = $_[1];

  my $lAgentLockFile = '/tmp/FromDualMySQLagent.lock';
  if ( $pLogLevel >= INFO ) { &log($pLogFile, INFO, '    ' . (caller(0))[3]); }

  # Check if lock file already exists and complain if yes
  if ( -e $lAgentLockFile ) {
    my $rc = 1314;
    if ( $pLogLevel >= WARN ) { &log($pLogFile, WARN, "    Agent Lock file $lAgentLockFile already exists."); }

    # Check the mtime of the lock file
    # If agent was running too long, kill it!
    my $mtime = (stat($lAgentLockFile))[9];
    my $now = time();

    # This is possibly too long for a local installation?
    my $timeout = 60;
    if ( $now-$mtime > $timeout ) {
      my $rc = 1315;
      &log($pLogFile, ERR, "    Agent was running for more than $timeout seconds. Killing it now! (rc=$rc)");
      `kill -9 $(cat $lAgentLockFile)`;
      &removeAgentLock;
    }
    else {
      my $rc = 1316;
      &log($pLogFile, WARN, "    Another agent is running. We will abort now! (rc=$rc)");
      exit $rc;
    }
  }

  # Create a new agent lock file
  my $ret = open(my $fh, '>', $lAgentLockFile);
  if ( ! defined($ret) ) {
    my $rc = 1310;
    &log($pLogFile, ERR, "  Cannot write to Lock File $lAgentLockFile: " . $! . "(rc=$rc)");
    exit $rc;
  }
  printf($fh "$$\n");
  close($fh);
}

sub removeAgentLock
{
  my $pLogLevel = $_[0];
  my $pLogFile  = $_[1];

  my $lAgentLockFile = '/tmp/FromDualMySQLagent.lock';
  if ( $pLogLevel >= INFO ) { &log($pLogFile, INFO, '    ' . (caller(0))[3]); }
  `rm -f $lAgentLockFile`;
}

1;
__END__
